package edu.bu.signstream.common;

public class SSMath {
    public static double clamp(double a, double b, double v) {
        return Math.max(a, Math.min(b, v));
    }

    /**
     * Linearly interpolates from [a, b] by fr in [0, 1] with no clamping of fr.
     */
    public static double lerpn(double a, double b, double fr) {
        return (a * (1.0 - fr)) + (b * fr);
    }
    public static double lerp(double a, double b, double fr) { return lerpn(a, b, clamp(0, 1, fr)); }

    /**
     * Linearly interpolates from [0, 1] by v in [a, b] with no clamping of v.
     */
    public static double ilerpn(double a, double b, double v) {
        return (v - a) / (b - a);
    }
    public static double ilerp(double a, double b, double v) { return ilerpn(a, b, clamp(a, b, v)); }

    /**
     * Projects v in [x, y] onto [a, b] with no clamping of v.
     */
    public static double projectn(double a, double b, double x, double y, double v) {
        return lerpn(a, b, ilerpn(x, y, v));
    }
    public static double project(double a, double b, double x, double y, double v) {
        return lerp(a, b, ilerp(x, y, v));
    }

    /**
     * Logarithmically interpolates from [a, b] by fr in [0, 1].
     * See https://www.desmos.com/calculator/ixdtn1t1ap
     */
    public static double loglerp(double A, double a, double b, double fr) {
        fr = clamp(0, 1, fr);
        return lerp(a, b, Math.log(1 - A * fr) / Math.log(1 - A));
    }
    public static double loglerp(double A, double fr) {
        return loglerp(A, 0, 1, fr);
    }
    public static double loglerpc(double A, double a, double b, double fr) { return loglerp(A, a, b, clamp(0, 1, fr)); }
    public static double loglerpc(double A, double fr) { return loglerp(A, clamp(0, 1, fr)); }

    /**
     * Logarithmically interpolates from [0, 1] by v in [a, b]
     * See https://www.desmos.com/calculator/ixdtn1t1ap
     */
    public static double logilerp(double A, double a, double b, double v) {
        v = clamp(a, b, v);
        return ilerp(a, b, (1 - Math.pow(1 - A, v)) / A);
    }
    public static double logilerp(double A, double v) {
        return logilerp(A, 0, 1, v);
    }
    public static double logilerpc(double A, double a, double b, double v) { return logilerp(A, a, b, clamp(a, b, v)); }
    public static double logilerpc(double A, double v) { return logilerp(A, clamp(0, 1, v)); }
}
