package edu.bu.signstream.common.tree;

import javax.swing.*;
import javax.swing.tree.*;
import java.util.Arrays;

public class Tree<T> {
    private final JTree tree;
    private final DefaultTreeModel model;

    public final Node<T> root;

    public Tree(Node<T> root) {
        this.root = root;
        model = new DefaultTreeModel(root.node());
        tree = new JTree(model);
        tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    }

    public JTree JTree() { return tree; }

    public Node<T> selected() {
        return tree.getSelectionPath() != null
            ? NodeWrapper.node((TreeNode)tree.getSelectionPath().getLastPathComponent())
            : null;
    }

    public void expand(Node<T> node) {
        var row = tree.getRowForPath(new TreePath(node.node().getPath()));
        tree.expandRow(row);
    }
    public void expandSubtree(Node<T> node) {
        node.dfs().forEach(this::expand);
    }
    public void expand() { expandSubtree(root); }

    public interface ReorderHandler<T> {
        boolean allowDrop(Node<T> node, Node<T> target, int index);
        void dropped(Node<T> node, Node<T> target, int index);
    }
    public void reorderable(ReorderHandler<T> handler) {
        tree.setDragEnabled(true);
        tree.setDropMode(DropMode.ON_OR_INSERT);
        tree.setTransferHandler(new TransferHandler() {
            @Override
            public int getSourceActions(JComponent c) { return COPY_OR_MOVE; }

            @Override
            public boolean canImport(TransferHandler.TransferSupport support) {
                if (support.getComponent() != tree) return false;
                if (!support.isDrop()) return false;
                support.setShowDropLocation(true);

                var loc = (JTree.DropLocation)support.getDropLocation();
                var path = loc.getPath();
                var row = tree.getRowForPath(path);
                var index = loc.getChildIndex();

                Node<T> target = NodeWrapper.node((TreeNode)path.getLastPathComponent());
                Node<T> node;
                try {
                    node = (Node<T>)support.getTransferable().getTransferData(target.getTransferDataFlavors()[0]);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }

                // Disallow moving node to itself
                if (Arrays.stream(tree.getSelectionRows()).anyMatch(r -> r == row))
                    return false;

                // Disallow moving node to its own child
                if (node.dfs().anyMatch(c -> c.equals(target)))
                    return false;

                return handler.allowDrop(node, target, index != -1 ? index : target.node().getChildCount());
            }

            @Override
            protected java.awt.datatransfer.Transferable createTransferable(JComponent c) {
                var paths = tree.getSelectionPaths();
                if (paths == null || paths.length == 0) return null;

                return NodeWrapper.node((TreeNode)paths[0].getLastPathComponent());
            }

            public boolean importData(TransferHandler.TransferSupport support) {
                if(!canImport(support)) return false;

                var loc = (JTree.DropLocation)support.getDropLocation();
                var path = loc.getPath();
                var index = loc.getChildIndex();

                Node<T> target = NodeWrapper.node((TreeNode)path.getLastPathComponent());
                Node<T> node;
                try {
                    node = (Node<T>)support.getTransferable().getTransferData(target.getTransferDataFlavors()[0]);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }

                var parent = node.parent();

                target.add(node, index);

                model.nodeStructureChanged(parent.node());
                model.nodeStructureChanged(target.node());

                handler.dropped(node, target, node.index());

                return true;
            }
        });
    }
}
