package edu.bu.signstream.common.util;

import edu.bu.signstream.time.Bound;
import edu.bu.signstream.time.Interval;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Optional;

// A data structure for keeping track of a large amount of
// intervals and determining if an interval overlaps with any of them.
public class IntervalSet<Key> {
    // Sorted list of overlapping intervals added so far, sorted by start time. No guarantee of non-overlapping intervals.
    private final ArrayList<Interval> intervals = new ArrayList<>();

    // Mapping from an added interval back to its key
    private final HashMap<Interval, Key> keys = new HashMap<>();

    // Return the indices of overlapping intervals
    private ArrayList<Integer> searchOverlap(Interval i, Bound b) {
    	ArrayList<Integer> list = new ArrayList<Integer>();
    	for (int x = 0; x < intervals.size(); x++) {
    		Interval o = intervals.get(x);
            
            if (o.overlaps(i, b)) {
            	list.add(x);
            }
    	}
    	
    	return list;
    }

    // Return the indices of containing intervals
    private ArrayList<Integer> searchContains(Interval i, Bound b) {
    	ArrayList<Integer> list = new ArrayList<Integer>();
    	for (int x = 0; x < intervals.size(); x++) {
    		Interval o = intervals.get(x);
            
            if (o.contains(i, b)) {
            	list.add(x);
            }
    	}
    	
    	return list;
    }

    // If overlapping, return the keys of the overlapping intervals.
    // Otherwise, add the interval to the set in order and return empty.
    public Optional<ArrayList<Key>> add(Key k, Interval i, Bound b) {
        if (intervals.isEmpty()) {
            intervals.add(i);
            keys.put(i, k);
            return Optional.empty();
        }
        
        ArrayList<Key> retKeys = new ArrayList<Key>();
    	
        for (int x = 0; x < intervals.size(); x++) {
	        	Interval o = intervals.get(x);
	        	Key okey = keys.get(o);
	
	        	if (o.overlaps(i, b)) {
	        		retKeys.add(okey);
	        	}
        }

        intervals.add(i);
        keys.put(i, k);
        intervals.sort(new Comparator<Interval>() {
			@Override
			public int compare(Interval interval0, Interval interval1) {
				if (interval0.t0.less(interval1.t0, b))
					return -1;
				else if (interval0.t0.greater(interval1.t0, b))
					return 1;
				else {
					if (interval0.t1.less(interval1.t1, b))
						return -1;
					else if (interval0.t1.greater(interval1.t1, b))
						return 1;
					else
						return 0;
				}
			}
        });

        if (retKeys.size() > 0)
        		return Optional.of(retKeys);
        else
        		return Optional.empty();
    }

    // If overlapping, return the keys of the overlapping intervals
    // Otherwise, return empty.
    public Optional<ArrayList<Key>> testOverlap(Interval i, Bound b) {
        if (intervals.isEmpty())
            return Optional.empty();
        
        ArrayList<Key> retKeys = new ArrayList<Key>();

        var idxList = searchOverlap(i, b);
        for (int x = 0; x < idxList.size(); x++) {
        	int idx = idxList.get(x);
            var interval = intervals.get(idx);
            retKeys.add(keys.get(interval));
        }

        return retKeys.isEmpty() ? Optional.empty() : Optional.of(retKeys);
    }

    // If contained, return the keys of the containing intervals
    // Otherwise, return empty.
    public Optional<ArrayList<Key>> testContains(Interval i, Bound b) {
        if (intervals.isEmpty())
            return Optional.empty();
        
        ArrayList<Key> retKeys = new ArrayList<Key>();

        var idxList = searchContains(i, b);
        for (int x = 0; x < idxList.size(); x++) {
        	int idx = idxList.get(x);
            var interval = intervals.get(idx);
            retKeys.add(keys.get(interval));
        }

        return retKeys.isEmpty() ? Optional.empty() : Optional.of(retKeys);
    }
    
    public ArrayList<Interval> getIntervals() {
    	return (ArrayList<Interval>) intervals.clone();
    }

    // very hacky - should find a better way to do this - even doing the for loop is getting around the 'final' for the arraylists
    public void setIntervals(ArrayList<Key> keys, ArrayList<Interval> intervals) {
    	this.intervals.clear();
    	this.keys.clear();
    	for (int i = 0; i < keys.size(); i++) {
    		Interval interval = intervals.get(i);
    		this.keys.put(interval, keys.get(i));
    		this.intervals.add(interval);
    	}
    }

    public ArrayList<String> getKeys() {
    	ArrayList<String> ret = new ArrayList<String>();
    	for (Interval i: intervals) {
    		ret.add(keys.get(i).toString());
    	}
    	return ret;
    }
}
