package edu.bu.signstream.common.util.https.requestBuilder;

import edu.bu.signstream.common.util.https.requestBuilder.wrappers.DataWrapper;
import edu.bu.signstream.common.util.https.requestBuilder.wrappers.FileWrapper;
import edu.bu.signstream.common.util.https.requestBuilder.wrappers.StringWrapper;
import org.apache.http.entity.ContentType;

import java.io.File;
import java.util.ArrayList;

/**
 * A RequestBuilder class that contains the information needed to build a request for a basic HTTP request: GET, POST, PUT, PATCH, DELETE
 */
public class RequestBuilder {
    private final RequestType requestType; // The type of request
    private final String url; // The url for the request
    private final String formLabel; // The form label for the request
    private final ArrayList<DataWrapper> wrappers; // The wrappers for the request
    private final ContentType contentType; // The content type for the request

    // GET request, DELETE request
    public RequestBuilder(String url, RequestType requestType) {
        if (requestType != RequestType.GET && requestType != RequestType.DELETE) {
            throw new IllegalArgumentException("Request type must be GET or DELETE");
        }

        this.url = url;
        this.formLabel = null; // GET and DELETE requests do not have a form label, wrappers, or content type
        this.wrappers = null;
        this.contentType = null;
        this.requestType = requestType;
    }

    // POST request, PUT request, PATCH request with File
    public RequestBuilder(String url, String formLabel, File file, ContentType contentType, RequestType requestType) {
        if (requestType != RequestType.POST && requestType != RequestType.PUT && requestType != RequestType.PATCH) {
            throw new IllegalArgumentException("Request type must be POST, PUT or PATCH");
        }

        this.url = url;
        this.formLabel = formLabel;
        this.wrappers = new ArrayList<>();
        this.wrappers.add(new FileWrapper(formLabel, file, contentType)); // We are only adding one file to the request for now
        this.contentType = contentType; // The content type for the request (ex: application/json)
        this.requestType = requestType; // The type of request (ex: POST)
    }

    /**
     * Same as above, but with a StringWrapper instead of a file wrapper
     * @param url The url for the request
     * @param formLabel The form label for the request
     * @param textBody The text body for the request
     * @param contentType The content type for the request
     * @param requestType The type of request
     */
    // POST request, PUT request, PATCH request with Text
    public RequestBuilder(String url, String formLabel, String textBody, ContentType contentType, RequestType requestType) {
        if (requestType != RequestType.POST && requestType != RequestType.PUT && requestType != RequestType.PATCH) {
            throw new IllegalArgumentException("Request type must be POST, PUT or PATCH");
        }

        this.url = url;
        this.formLabel = formLabel;
        this.wrappers = new ArrayList<>();
        this.wrappers.add(new StringWrapper(formLabel, textBody, contentType));
        this.contentType = contentType;
        this.requestType = requestType;
    }

    public RequestType getRequestType() {
        return this.requestType;
    }

    public String getUrl() {
        return this.url;
    }

    public String getFormLabel() {
        return this.formLabel;
    }

    public ArrayList<DataWrapper> getWrappers() {
        return this.wrappers;
    }

    public ContentType getContentType() {
        return this.contentType;
    }
}
