package edu.bu.signstream.grepresentation.fields.graphField;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Line2D;
import java.awt.geom.QuadCurve2D;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import edu.bu.signstream.common.SSMath;
import edu.bu.signstream.time.Bound;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;
import org.apache.commons.math3.analysis.interpolation.SplineInterpolator;
import org.apache.commons.math3.analysis.polynomials.PolynomialSplineFunction;

import edu.bu.signstream.common.util.SegmentGraphReprUtil;
import edu.bu.signstream.common.util.vo.ss3.db.SS3Track;
import edu.bu.signstream.grepresentation.fields.GraphFileRow;
import edu.bu.signstream.grepresentation.fields.nonManualField.NonManualField;
import edu.bu.signstream.grepresentation.fields.presentation.PresentationField;
import edu.bu.signstream.grepresentation.view.main.SignStreamSegmentPanel;

public class GraphField extends NonManualField {
    private List<GraphFileRow> data;
    protected boolean maximized = false;

	public GraphField(SS3Track track, SignStreamSegmentPanel segmentPanel, PresentationField parentField) {
		super(segmentPanel, parentField);
		this.ss3Track = track;
	}

	// override getFieldID and return tmp ID (to get SignSteamSegmentPanel to work
	// where indicated) until SS3Track is ready
	@Override
	public String getFieldID() {
		return ss3Track.getId();
	}

	public void setRawGraphList(List<GraphFileRow> rowList) {
		this.data = rowList;
	}

	@Override
	public void paint(Graphics2D g2d, double Y, Font font) {
	    if (data.isEmpty()) return;

		// copy the graphics passed in so we can set different rending hints specifically for this field
		Graphics2D g2 = (Graphics2D) g2d.create();		
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);

		// Min and max wrapper y coordinates
		int wrapperHalfHeight = SegmentGraphReprUtil.getCorrectCellHeight(this) / 2;
		int offset = maximized ? 16 : 2;
		Y = Y - wrapperHalfHeight / 2.0 - offset;
		double wrapperMin = Y - wrapperHalfHeight + offset;
		double wrapperMax = Y + wrapperHalfHeight - offset;

		// Convert data to coordinates
        var coords = data.stream()
            .map(r -> {
                // Convert the time into an absolute coordinate
                var x = r.time.absolute().as(TimeUnit.COORDINATE);
                // Project the [-1, 1] normalized value onto the wrapper coordinate space,
                // reversing the min and max since coordinates increase downwards.
                var y = SSMath.projectn(wrapperMax, wrapperMin, -1, 1, r.value);
                return new Point(x, y);
            })
            .collect(Collectors.toList());

		// styles for the graphics objects
		float px1 = 1f;
		float px2 = 2f;
		float[] dashingPattern = { px2, px2 };
		Color medianColor = Color.GRAY;
		Color graphColor = Color.RED;

        var xmin = coords.stream()
            .map(p -> p.x)
            .min(Double::compare).get();
        var xmax = coords.stream()
            .map(p -> p.x)
            .max(Double::compare).get();

        // draw median as a dotted line from xmin to xmax
        double ymid = (wrapperMin + wrapperMax) / 2;
        g2.setColor(medianColor);
        g2.setStroke(new BasicStroke(px1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, px1, dashingPattern, px2));
        g2.draw(new Line2D.Double(xmin, ymid, xmax, ymid));

        // Create a list of x and y coordinates, removing any
        // duplicate xs since the spline interpolation requires
        // monotonically increasing values.
        var xs = new ArrayList<Double>();
        var ys = new ArrayList<Double>();
        Point prev = null;
        for (var p : coords) {
            if (prev != null && p.x <= prev.x)
                continue;
            prev = p;

            xs.add(p.x);
            ys.add(p.y);
        }

        // Calculate the spline function for rendering
        SplineInterpolator si = new SplineInterpolator();
        PolynomialSplineFunction psf = si.interpolate(
            xs.stream().mapToDouble(Double::doubleValue).toArray(),
            ys.stream().mapToDouble(Double::doubleValue).toArray());

        g2.setStroke(new BasicStroke(px2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
        g2.setColor(graphColor);

        // Paint the lines between each point
        double lastDiff = 0.0;
        for (int i = 0; i < xs.size() - 1; i++) {
            var xi = xs.get(i);
            var xj = xs.get(i + 1);
            var yi = ys.get(i);
            var yj = ys.get(i + 1);

            // Don't paint if the line isn't overlapping with the visible interval
            var ti = TimeRange.ABSOLUTE.time(Math.round(xi), TimeUnit.COORDINATE);
            var tj = TimeRange.ABSOLUTE.time(Math.round(xj), TimeUnit.COORDINATE);
            if (!ti.to(tj).overlaps(SegmentGraphReprUtil.visibleInterval(), Bound.INCLUSIVE))
                continue;

            // Don't paint if the x coordinate jumps by more than 200% compared to the last point
            double currDiff = Math.abs(xj - xi);
            if (currDiff / lastDiff < 2)
                precisionPaint(1, 0.5, g2, psf, xi, yi, xj, yj);
            lastDiff = currDiff;
        }
	}
	
	/**
	 * Precision corresponds to how many points to interpolate between the real (frame) points
	 * Following the expression: numPoints = Math.pow(2, p) - 1, where p (precision) >= 1 (can we go to zero? for no precision?))
	 * Recursive function which decreases 
	 * @param precision
	 */
	private void precisionPaint(int precision, double tension, Graphics2D g2, PolynomialSplineFunction psf, 
			double xa, double ya, double xc, double yc) {
		
		// base case - done interpolating extra points, and stop painting
		if (precision <= 0)
			return;

		// interpolated value b (between [a, c])
		double xb = (xa + xc) / 2;		
		double yb = psf.value(xb);

	    double distAB = Math.sqrt(Math.pow( (xb - xa), 2) + Math.pow( (yb - ya), 2) );
		double distBC = Math.sqrt(Math.pow( (xc - xb), 2) + Math.pow( (yc - yb) , 2) );
		
	    double fa = tension * distAB / (distAB+distBC);   // scaling factor for triangle Ta
	    double fb = tension * distBC / (distAB+distBC);   // ditto for Tb, simplifies to fb=t-fa
	    
	    double xACwidth = (xc - xa);
	    double yACheight = (yc - ya);
	    
	    double cp1x = xb - fa * xACwidth;
	    double cp1y = yb - fa * yACheight;
	    double cp2x = xb + fb * xACwidth;
	    double cp2y = yb + fb * yACheight;
	    
////	// A. method: curve using midpoints as control points (known wrong)
////	
////	shape.setCurve(beginX, beginRelY, 
////							cp1x, cp1y, 
////							cp2x, cp2y, 
////							endX, endRelY);
////	g2.draw(shape);				
//	
//	
//	    // B. method: 3 straight lines through point1, control1, control2, point2	
//	    g2.drawLine((int)xa, (int)ya, 
//	    		(int)xb, (int)yb);
//	    g2.drawLine((int)xb, (int)yb, 
//	    		(int)xc, (int)yc);

	    // C. method: calculate control points from the polynomial function - with equal derivatives on both sides, then graph curves
		// http://scaledinnovation.com/analytics/splines/aboutSplines.html
	    QuadCurve2D  quadShape = new QuadCurve2D.Double();
		quadShape.setCurve(xa, ya, 
								cp1x, cp1y, 
								xb, yb);
		g2.draw(quadShape);		
		quadShape.setCurve(xb, yb, 
								cp2x, cp2y, 
								xc, yc);
		g2.draw(quadShape);	
	    

		precision--;
		precisionPaint(precision, tension, g2, psf, xa, ya, xb, yb);
		precisionPaint(precision, tension, g2, psf, xb, yb, xc, yc);
	}

	public boolean isMaximized() {
		return maximized;
	}

	public void toggleMaximized() {
		maximized = !maximized;
	}
}

class Point {
	public final double x;
	public final double y;

	public Point(double x, double y) {
		this.x = x;
		this.y = y;
	}

    @Override
	public String toString() {
		return "Point{x=" + x + " y=" + y + "}";
	}
}