package edu.bu.signstream.grepresentation.fields.palettes.helpers;

import edu.bu.signstream.common.util.ErrorMessages;
import edu.bu.signstream.common.util.Pair;
import edu.bu.signstream.common.util.vo.Constants;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3CodingScheme;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.templates.*;
import edu.bu.signstream.grepresentation.requestsHandler.SS3Singleton;
import edu.bu.signstream.ui.SS3SignStreamApplication;

import javax.swing.*;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.logging.Level;

// Policy: Every public method in this class MUST return a value.

public class PaletteInformationHelper {
    private final SS3CodingScheme scheme;
    private final String defaultPaletteMenuID;
    private String currentTemplateID; // This refers to the default template from the XML (like ASL)
    private ArrayList<PaletteGroup> paletteGroups;
    private ArrayList<CSHandShapePalette> handShapePalettes; // This is the list of palettes that are in the current template (like ASL template)
    private String currentDomPaletteID; // This refers to the current selected palette from the combobox (like ASL - ALL or ASL - Numbers, etc)
    private String currentNDomPaletteID; // This refers to the current selected palette from the combobox (like ASL - ALL or ASL - Numbers, etc)
    private HashMap<String, PaletteReference> paletteReferences;

    private boolean inconsistentXML = false;

    public PaletteInformationHelper(SS3CodingScheme scheme, String defaultPaletteMenuID) {
        this.scheme = scheme;
        this.defaultPaletteMenuID = defaultPaletteMenuID;
        this.currentTemplateID = this.getDefaultTemplateID();
    }

    /**
     * This method will get the default template ID based on what is marked in the XML in the PALETTE_MENUS section in the
     * defCodingScheme.xml. Depending on the default tag in the PALETTE_REFS, the default template ID will be set.
     * If there are multiple groups, we do parse all groups and make sure all default template IDs are the same so that the
     * template IDs are consistent. This is subject to change, however.
     * @return The default template ID
     */
    private String getDefaultTemplateID() {
        this.paletteGroups = this.getDefaultPaletteMenu().getPaletteGroups(); // Based on whether this is a gloss or gesture, we get the palette groups. That's what the getDefaultPaletteMenu is for

        for (PaletteGroup paletteGroup : this.paletteGroups) {
            for (PaletteReference paletteReference : paletteGroup.getPaletteReferences()) {
                if (paletteReference.getIsDomDefault() || paletteReference.getIsNDomDefault()) { // We check if this paletteReference has the default tag, if so, that will allow us to set the default templateID,
                    if (this.currentTemplateID == null) {
                        this.currentTemplateID = paletteReference.getTemplateId();
                    } else if (!paletteReference.getTemplateId().equals(this.currentTemplateID)) { // If it is a default tag, we check if the template ID is the same as the default template ID. If not, we throw an error as this means the defCodingScheme.xml is in an inconsistent state.
                        SS3Singleton.getErrorMessages().showErrorMessage(ErrorMessages.INCONSISTENT_XML_TEMPLATE_IDS);

                        this.inconsistentXML = true;

                        SS3SignStreamApplication.logger.log(Level.SEVERE,
                                String.format("Palette menu %s has inconsistent template IDs. Please check the PALETTE GROUPS in the defCodingScheme.xml file.", this.defaultPaletteMenuID));

                        return null;
                    }
                }
            }
        }

        return this.currentTemplateID;
    }

    /**
     * @return The palette menu that makes the default menu ID set in the defCodingScheme.xml
     */
    private PaletteMenu getDefaultPaletteMenu() {
        ArrayList<PaletteMenu> paletteMenus = this.scheme.getPaletteMenus();

        for (PaletteMenu paletteMenu : paletteMenus) {
            if (paletteMenu.getId().equals(this.defaultPaletteMenuID)) {
                return paletteMenu;
            }
        }

        SS3SignStreamApplication.logger.log(SS3SignStreamApplication.logger.getLevel(), String.format("Palette menu %s does not exist in the defCodingScheme.xml.", this.defaultPaletteMenuID));
        SS3Singleton.getErrorMessages().showErrorMessage(ErrorMessages.PALETTE_MISSING);

        return null;
    }


    /**
     * This method will add the templates into the combo-box based on the defCodingScheme.xml
     */
    public ArrayList<Pair<String, Boolean>> getTemplateData() {
        ArrayList<Pair<String, Boolean>> templateNames = new ArrayList<>();

        if (this.currentTemplateID == null) { // This means the XML is in an inconsistent state. We do not want to populate the combobox if this is the case.
            return null;
        }

        ArrayList<CSHandShapeTemplate> templates = this.scheme.getHandShapeTemplates().getTemplatesAsList();
        templates.sort(Comparator.comparingInt(CSHandShapeTemplate::getOrder)); // Sort the templates based on the order attribute in the defCodingScheme.xml

        for (CSHandShapeTemplate template : templates) {
            if (template.getId().equals(this.currentTemplateID)) { // We check if the current template is the default template, then when the dialog opens, that template should be selected by default.
                templateNames.add(Pair.of(template.getName(), true));
            } else {
                templateNames.add(Pair.of(template.getName(), false)); // False because it is not the default template
            }
        }

        return templateNames;
    }


    /**
     * @param paletteGroups The palette groups that are in the defCodingScheme.xml
     * @param id The ID of the palette group that we want to get the palette references for (Whether you want handshapes or gestures, depending on where you're calling from)
     * @return Return the references as a HashMap
     */
    private HashMap<String, PaletteReference> getPaletteReferences(ArrayList<PaletteGroup> paletteGroups, String id) {
        for (PaletteGroup paletteGroup : paletteGroups) {
            if (paletteGroup.getId().equals(id)) {
                return paletteGroup.getPaletteReferencesAsHashMap();
            }
        }

        return null;
    }


    /**
     * @param templateID The template ID that we want to get the palette references for, in this case, it will be the default template ¬ID (calling from the constructor)
     * @param paletteGroupID The palette group ID that we want to get the palette references for. This might be Handshapes or Gestures, depending on what we are trying to access. TODO: Hasnain, constructor is subject to change based on what we want to populate.
     */
    public ArrayList<Pair<String, Boolean>> getPaletteItemGroupOptions(String templateID, String paletteGroupID, boolean isDomSelected, boolean isNDomSelected) {
        ArrayList<Pair<String, Boolean>> paletteItemGroupOptions = new ArrayList<>();
        this.paletteReferences = this.getPaletteReferences(this.paletteGroups, paletteGroupID);

        CSHandShapeTemplate defaultTemplate = this.scheme.getHandShapeTemplates().getTemplates().get(templateID); // Get the default template based on the template ID
        String palettePrefix = defaultTemplate.getLabel(); // For example: "ASL" for "American Sign Language" ~ Straight from the defCodingScheme.xml
        ArrayList<CSHandShapePalette> handShapePalettes = new ArrayList<>(defaultTemplate.getPalettes()); // Get the palettes for the default template (for example, ASL - ALL or ASL - Numbers)

        // Filter out extra handShapePalettes that don't have IDs in the paletteReferences
        handShapePalettes.removeIf(palette -> !(this.paletteReferences.containsKey(palette.getId()))); // This is to make sure that the palette references select the specific palettes we want from the template as all of them may not be used, so we need to filter them out.

        handShapePalettes.sort((o1, o2) -> {
            assert this.paletteReferences != null;
            int o1Order = Integer.parseInt(this.paletteReferences.get(o1.getId()).getOrder());
            int o2Order = Integer.parseInt(this.paletteReferences.get(o2.getId()).getOrder());

            return o1Order - o2Order;
        });

        this.handShapePalettes = handShapePalettes;

        for (CSHandShapePalette palette : this.handShapePalettes) {
            if (isDomSelected) {
                if (this.paletteReferences.get(palette.getId()).getIsDomDefault()) {
                    this.currentDomPaletteID = palette.getId();
                    paletteItemGroupOptions.add(Pair.of(palettePrefix + " - " + palette.getName(), true));
                    continue;
                }
            } else if (isNDomSelected) {
                if (this.paletteReferences.get(palette.getId()).getIsNDomDefault()) {
                    this.currentNDomPaletteID = palette.getId();
                    paletteItemGroupOptions.add(Pair.of(palettePrefix + " - " + palette.getName(), true));
                    continue;
                }
            }

            paletteItemGroupOptions.add(Pair.of(palettePrefix + " - " + palette.getName(), false));
        }

        return paletteItemGroupOptions;
    }

    public String getPaletteNameFromID(String id) {
        // Get prefix from template
        CSHandShapeTemplate defaultTemplate = this.scheme.getHandShapeTemplates().getTemplates().get(this.currentTemplateID);
        String palettePrefix = defaultTemplate.getLabel();

        // Get palette name from palette ID
        for (CSHandShapePalette palette : this.handShapePalettes) {
            if (palette.getId().equals(id)) {
                return palettePrefix + " - " + palette.getName();
            }
        }

        return null;
    }

    public void setPaletteItemGroupComboBox(JComboBox<String> paletteItemGroupComboBox, String templateID, String paletteGroupID, String paletteID) {
        this.paletteReferences = this.getPaletteReferences(this.paletteGroups, paletteGroupID);

        CSHandShapeTemplate defaultTemplate = this.scheme.getHandShapeTemplates().getTemplates().get(templateID); // Get the default template based on the template ID
        String palettePrefix = defaultTemplate.getLabel();

        for (CSHandShapePalette palette : this.handShapePalettes) {
            paletteItemGroupComboBox.addItem(palettePrefix + " - " + palette.getName());

            if (palette.getId().equals(paletteID)) {
                paletteItemGroupComboBox.setSelectedItem(palettePrefix + " - " + palette.getName());
            }
        }
    }

    public String getCurrentDomPaletteID() {
        return this.currentDomPaletteID;
    }

    public String getCurrentNDomPaletteID() {
        return this.currentNDomPaletteID;
    }

    public void setCurrentNDomPaletteID(String currentNDomPaletteID) {
        this.currentNDomPaletteID = currentNDomPaletteID;
    }

    public void setCurrentDomPaletteID(String currentDomPaletteID) {
        this.currentDomPaletteID = currentDomPaletteID;
    }

    public boolean getInconsistentXML() {
        return this.inconsistentXML;
    }

    public String getCurrentTemplateID() {
        return this.currentTemplateID;
    }

    public ArrayList<CSHandShapePalette> getHandShapePalettes() {
        return this.handShapePalettes;
    }
}
