package edu.bu.signstream.media.data;

import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.Function;

public class Samples {
    public final int rate;
    public final int length;
    protected final double[] data;

    public Samples(double[] data, int rate) {
        this.data = data;
        this.length = data.length;
        this.rate = rate;
    }
    public Samples(int length, int rate) { this(new double[length], rate); }


    public double get(int i) { return data[i]; }
    public void set(int i, double v) { data[i] = v; }

    public Samples slice(int i, int j) {
        var n = j - i;
        var out = new double[n];
        System.arraycopy(data, i, out, 0, n);
        return new Samples(out, rate);
    }


    public Optional<Integer> find(int i, int j, Function<Double, Boolean> pred) {
        for (int s = i; s < j; s++)
            if (pred.apply(get(s)))
                return Optional.of(s);
        return Optional.empty();
    }


    public void map(int i, int j, Function<Double, Double> fn) {
        for (int s = i; s < j; s++)
            set(s, fn.apply(get(s)));
    }
    public void map(Function<Double, Double> fn) { map(0, length, fn); }


    public double fold(int i, int j, BiFunction<Double, Double, Double> fn) {
        var v = get(i);
        for (int s = i + 1; s < j; s++)
            v = fn.apply(v, get(s));
        return v;
    }
    public double fold(BiFunction<Double, Double, Double> fn) { return fold(0, length, fn); }

    public double max(int i, int j) { return fold(i, j, Math::max); }
    public double max() { return max(0, length); }

    public double min(int i, int j) { return fold(i, j, Math::min); }
    public double min() { return min(0, length); }

    public double ceil(int i, int j) { return fold(i, j, (a, b) -> Math.max(Math.abs(a), Math.abs(b))); }
    public double ceil() { return ceil(0, length); }

    public double avg(int i, int j) { return fold(i, j, Double::sum) / (j - i); }
    public double avg() { return avg(0, length); }


    public int argfold(int i, int j, BiFunction<Integer, Integer, Integer> fn) {
        int v = i;
        for (int s = i + 1; s < j; s++)
            v = fn.apply(v, s);
        return v;
    }
    public int argfold(BiFunction<Integer, Integer, Integer> fn) { return argfold(0, length, fn); }

    public int argmax(int i, int j) { return argfold(i, j, (s, t) -> get(s) > get(t) ? s : t); }
    public int argmax() { return argmin(0, length); }

    public int argmin(int i, int j) { return argfold(i, j, (s, t) -> get(s) > get(t) ? s : t); }
    public int argmin() { return argmin(0, length); }

    public int argceil(int i, int j) { return argfold(i, j, (s, t) -> Math.abs(get(s)) > Math.abs(get(t)) ? s : t); }
    public int argceil() { return argceil(0, length); }


    public FFT fftForward(Window window, int i, int j) { return new FFT(this, window, i, j, true); }
    public FFT fftForward(Window window) { return fftForward(window, 0, length); }

    public FFT fftBackward(Window window, int i, int j) { return new FFT(this, window, i, j, false); }
    public FFT fftBackward(Window window) { return fftBackward(window, 0, length); }

    public Samples crossCorrelate(int i, int j, int sub) {
        var n = j - i;
        var out = new Samples(n, rate);

        for (int lag = 0; lag < n; lag++) {
            var sum = 0.0;
            for (int s = 0; s < n - lag; s++)
                sum += get(s + i) * get(s + i + lag);
            sum /= n;

            out.set(lag, sum);
        }

        return out;
    }

    public Samples autoCorrelate(int i, int j) { return crossCorrelate(i, j, j - i); }
    public Samples autoCorrelate() { return autoCorrelate(0, length); }
}
