package edu.bu.signstream.media.gstreamer;

import java.net.URI;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import edu.bu.signstream.time.*;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.media.*;
import edu.bu.signstream.media.gstreamer.util.Executor;
import org.freedesktop.gstreamer.*;
import org.freedesktop.gstreamer.event.SeekEvent;
import org.freedesktop.gstreamer.event.SeekFlags;
import org.freedesktop.gstreamer.event.SeekType;

import javax.swing.*;
import javax.swing.Timer;

public class GstMediaPlayer implements MediaPlayer {
    // GStreamer
    final Pipeline pipeline;
    final Bus bus;

    // Media
    private final ArrayList<GstMedia> media = new ArrayList<>();
    private final HashMap<Media, Duration> offsets = new HashMap<>();

    // State
    private PlayerState state;
    private PlayerState prevState = null;
    public Duration duration = new Duration(0, TimeUnit.NANOSECOND);

    // Listeners
    private final List<Consumer<Time>> timeListeners = new ArrayList<>();
    private final List<Consumer<Boolean>> stateListeners = new ArrayList<>();
    private final List<Consumer<Double>> rateListeners = new ArrayList<>();

    // Executors
    private final Executor exec;
    private final Executor gst;

    public GstMediaPlayer() {
        if (!Gst.isInitialized()) {
            System.out.println("M: INIT GStreamer");
            Gst.init();
            System.out.println("M: INITIALIZED " + Gst.getVersionString());
        }

        exec = new Executor("media player");
        gst = new Executor(Gst.getExecutor(), "gstreamer service thread");

        System.out.println("M: PLUGINS:");
        for (var p : Registry.get().getPluginList())
            System.out.println(" - " + p);

        pipeline = new Pipeline();
        bus = pipeline.getBus();

        bus.connect((Bus.INFO)(arg0, arg1, arg2) ->
            System.err.println("M: <main:info> " + arg0.getName() + " [" + arg1 + "]: " + arg2));

        bus.connect((Bus.ERROR)(arg0, arg1, arg2) ->
            System.err.println("M: <main:error> " + arg0.getName() + " [" + arg1 + "]: " + arg2));

//        bus.connect((Bus.MESSAGE)(arg0, arg1) ->
//            System.out.println("M: <main:msg> " + arg1.getSource().getName() + " [" + arg1.getType() + "]: " + arg1.getStructure()));

        this.state = this.new NotReady();
    }

    @Override
    public List<Media> media() {
        return Collections.unmodifiableList(media);
    }

    @Override
    public Optional<Media> media(URI uri) {
        return media.stream()
            .filter(m -> m.uri().equals(uri))
            .map(m -> (Media)m)
            .findFirst();
    }

    @Override
    public List<Video> videos() {
        return media.stream()
            .filter(m -> m instanceof GstVideo)
            .sorted(Comparator.comparing(v -> ((GstVideo) v).id()))
            .map(v -> (Video)v)
            .collect(Collectors.toUnmodifiableList());
    }

    @Override
    public Optional<Video> video(URI uri) {
        return media.stream()
            .filter(m -> m instanceof GstVideo)
            .filter(m -> m.uri().equals(uri))
            .map(v -> (Video)v)
            .findFirst();
    }

    @Override
    public List<Audio> audios() {
        return media.stream()
            .filter(m -> m instanceof GstAudio)
            .sorted(Comparator.comparing(a -> ((GstAudio) a).id()))
            .map(a -> (Audio)a)
            .collect(Collectors.toUnmodifiableList());
    }

    @Override
    public Optional<Audio> audio(URI uri) {
        return media.stream()
            .filter(m -> m instanceof GstAudio)
            .filter(m -> m.uri().equals(uri))
            .map(v -> (Audio)v)
            .findFirst();
    }

    @Override
    public Video addVideo(URI uri, int order) {
        System.out.println("M: ADD " + uri);

        var video = new GstVideo(uri, order, this);
        media.add(video);
        offset(video, Duration.ZERO);
        video.display();

        return video;
    }

    @Override
    public Audio addAudio(URI uri, int order) {
        System.out.println("addAudio(" + uri + ")");

        var audio = new GstAudio(uri, order, this);
        media.add(audio);
        offset(audio, Duration.ZERO);
        audio.display();

        return audio;
    }

    @Override
    public void remove(Media m) {
        System.out.println("M: REMOVE " + m.uri());
        gst.run(((GstMedia) m)::close);
        media.remove(m);
    }


    @Override
    public void offset(Media m, Duration d) {
        m.range().offset(d);
        offsets.put(m, d);

        System.out.println("M: OFFSET " + m.uri() + " " + d);

        duration = media.stream()
            .map(f -> f.duration()
                .add(offsets.get(f))
                .subtract(new Duration(1, f.range().unit())))
            .reduce(Duration::max)
            .map(max -> max.convert(TimeUnit.NANOSECOND))
            .orElseThrow();

        System.out.println("M: OFFSET DURATION " + duration);
    }


    @Override
    public void preroll() {
        if (prerolled())
            return;

        System.out.println("M: PREROLL");
        state.preroll();
    }

    @Deprecated
    public boolean prerolled() {
        return !(state instanceof NotReady);
    }

    @Override
    public void play() {
        if (playing() || time().equals(end()))
            return;

        System.out.println("M: PLAY");
        state.play();
    }

    @Override
    public void pause() {
        if (!playing())
            return;

        System.out.println("M: PAUSE");
        state.pause();
    }

    @Override
    public boolean playing() {
        return state.segment().playing;
    }

    @Override
    public void rate(double rate) {
        System.out.println("M: RATE " + rate);
        state.seek(new Segment(time(), range(), rate, playing()));
    }

    @Override
    public double rate() {
        return state.segment().rate;
    }

    @Override
    public void seek(Time time) {
        System.out.println("M: SEEK " + time);
        state.seek(new Segment(time, time.to(end()), rate(), playing()));
    }

    @Override
    public void seek(Interval interval) {
        System.out.println("M: SEEK " + interval);
        state.seek(new Segment(interval.t0, interval, rate(), playing()));
    }

    @Override
    public void seek(Interval interval, double rate) {
        System.out.println("M: SEEK " + interval + " (x" + rate + ")");
        state.seek(new Segment(interval.t0, interval, rate, playing()));
    }

    @Override
    public Time time() {
        return state.segment().time;
    }

    @Override
    public Interval range() {
        return state.segment().range;
    }

    @Override
    public Duration duration() {
        return duration;
    }

    @Override
    public Time start() {
        return TimeRange.ABSOLUTE.zero();
    }

    @Override
    public Time end() {
        return start().add(duration);
    }

    @Override
    public Interval interval() {
        return start().to(end());
    }


    @Override
    public void addTimeListener(Consumer<Time> fn) {
        timeListeners.add(fn);
    }
    @Override
    public void removeTimeListener(Consumer<Time> fn) {
        timeListeners.remove(fn);
    }
    private void fireTimeListeners(Time t) {
        if (t.greaterThanOrEq(start())) {
            var clamped = t.max(start()).min(end()).absolute();
//            System.out.println("M: FIRE TIME " + t + " (" + clamped + ")");
            for (var l : timeListeners)
                l.accept(clamped);
        } else {
            System.out.println("M: DROP TIME " + t);
        }
    }

    @Override
    public void addStateListener(Consumer<Boolean> fn) {
        stateListeners.add(fn);
    }
    @Override
    public void removeStateListener(Consumer<Boolean> fn) {
        stateListeners.remove(fn);
    }
    private void fireStateListeners(boolean state) {
//        System.out.println("M: FIRE STATE " + (state ? "PLAYING" : "PAUSED"));
        for (var l : stateListeners)
            l.accept(state);
    }

    @Override
    public void addRateListener(Consumer<Double> fn) {
        rateListeners.add(fn);
    }
    @Override
    public void removeRateListener(Consumer<Double> fn) {
        rateListeners.remove(fn);
    }
    private void fireRateListeners(double rate) {
//        System.out.println("M: FIRE RATE " + rate);
        for (var l : rateListeners)
            l.accept(rate);
    }

    @Override
    public void close() {
        System.out.println("M: CLOSE");

        for (var m : media)
            m.hide();

        state.close();
    }

    private long pos() {
        long pos = -1;
        for (var m : media) {
            var offset = offsets.get(m).as(TimeUnit.NANOSECOND, Rounding.NEAREST);
            var tsOffset = (long)m.sink().get("ts-offset");

            long mpos = m.pos();
//            System.out.println("M: POS " + m.uri() + ": " + mpos + " (+" + offset + ") (+ts" + tsOffset + ")" + " IN " + m.sink().getState());
            if (mpos == -1)
                new RuntimeException("-1 pos").printStackTrace();

            pos = Math.max(pos, mpos + offset - tsOffset);
        }

        return pos;
    }

    private static class Segment {
        public final Time time;
        public final Interval range;
        public final double rate;
        public final boolean playing;

        public Segment(Time time, Interval range, double rate, boolean playing) {
            this.time = time;
            this.range = range;
            this.rate = rate;
            this.playing = playing;
        }

        @Override
        public String toString() {
            return "Segment{" + time + " (" + range + ") x" + rate + " " + (playing ? "PLAYING" : "PAUSED") + "}";
        }
    }

    private abstract class PlayerState {
        protected Segment segment;

        public PlayerState(Segment segment) {
            this.segment = segment;
        }

        public Segment segment() { return segment; }
        public void segment(Segment segment) {
            this.segment = segment;
            fireTimeListeners(segment.time);
            fireRateListeners(segment.rate);
            fireStateListeners(segment.playing);
        }

        // Controls //

        public void preroll() {}

        public void play() {
            transition(this, () -> new Playing(segment()));
        }
        public void pause() {
            transition(this, () -> new Paused(segment()));
        }
        public void seek(Segment segment) {
            transition(this, () -> new Seeking(segment(), segment));
        }
        public void close() {
            transition(this, () -> new NotReady());
        }
    }

    class NotReady extends PlayerState {
        public NotReady() {
            super(new Segment(TimeRange.ABSOLUTE.zero(), TimeRange.ABSOLUTE.zero().to(TimeRange.ABSOLUTE.zero()), 1, false));

            for (var m : new ArrayList<>(media))
                remove(m);

            gst.run(() -> pipeline.setState(State.NULL));
        }

        @Override
        public void preroll() {
            var state = gst.call(pipeline::getState);
            expect(state == State.NULL, "Preroll in invalid pipeline state " + state);

            System.out.println("M: PREROLLING STATE=" + state);

            transition(this, () -> new Paused(new Segment(start(), range(), 1.0, false)));
        }

        @Override public void play() {}
        @Override public void pause() {}
        @Override public void seek(Segment segment) {}
        @Override public void close() {}

        @Override public String toString() { return "NotReady{}"; }
    }

    class Paused extends PlayerState {
        public Paused(Segment segment) {
            super(segment);

            var state = gst.call(pipeline::getState);
            expect(state == State.PAUSED || state == State.PLAYING || state == State.NULL,
                "Pause in invalid pipeline state " + state);

            System.out.println("M: PAUSING STATE=" + state);

            if (state == State.NULL) {
                gst.run(pipeline::pause);
            } else if (playing()) {
                var latch = new CountDownLatch(1);
                pipeline.getBus().connect(new Bus.STATE_CHANGED() {
                    @Override
                    public void stateChanged(GstObject source, State old, State current, State pending) {
                        if (source != pipeline || current != State.PAUSED)
                            return;

                        pipeline.getBus().disconnect(this);

                        latch.countDown();
                    }
                });
                gst.call(pipeline::pause);
                await(latch);
            }

            state = gst.call(pipeline::getState);
            expect(state == State.PAUSED, "Post-pause in invalid pipeline state " + state);

            var time = TimeRange.ABSOLUTE.time(gst.call(GstMediaPlayer.this::pos), TimeUnit.NANOSECOND);
            segment(new Segment(time, range(), rate(), false));
        }

        @Override public void pause() {}

        @Override public String toString() { return "Paused{}"; }
    }

    class Playing extends PlayerState {
        private static final int POLL_RATE = 30;

        private final AtomicBoolean eos = new AtomicBoolean(false);
        private final Bus.EOS eosListener;

        private final Timer timer = new Timer(POLL_RATE, e -> {
            var time = TimeRange.ABSOLUTE.time(gst.call(GstMediaPlayer.this::pos), TimeUnit.NANOSECOND);
            if (!time().equals(time)) {
                segment(new Segment(time, range(), rate(), playing()));
            }
        });

        public Playing(Segment segment) {
            super(segment);

            eosListener = source -> {
                eos.set(true);
                cleanup();

                var state = pipeline.getState();
                expect(state == State.PLAYING, "Playing EOS in invalid state " + state);
                System.out.println("M: EOS STATE=" + state);

                var target = new Segment(range().t1, interval(), rate(), false);
                transitionAsync(this, () -> {
                    new Paused(segment()); // pause without changing states
                    return new Seeking(target, target);
                });
            };
            bus.connect(eosListener);

            if (!playing()) {
                var state = gst.call(pipeline::getState);
                expect(state == State.PAUSED, "Play in invalid state " + state);
                System.out.println("M: PLAYING STATE=" + state);

                var latch = new CountDownLatch(1);
                bus.connect(new Bus.STATE_CHANGED() {
                    @Override
                    public void stateChanged(GstObject source, State old, State current, State pending) {
                        if (source != pipeline || current != State.PLAYING)
                            return;

                        bus.disconnect(this);
                        latch.countDown();
                    }
                });
                gst.call(pipeline::play);
                await(latch);
            }

            var state = gst.call(pipeline::getState);
            expect(state == State.PLAYING, "Post-Play in invalid state " + state);

            if (!eos.get()) {
                segment(new Segment(time(), range(), rate(), true));
                timer.start();
            }
        }

        private void cleanup() {
            timer.stop();
            bus.disconnect(eosListener);
        }

        @Override public void play() {}

        @Override
        public void pause() {
            if (eos.get())
                return;

            cleanup();
            super.pause();
        }

        @Override
        public void seek(Segment segment) {
            if (eos.get())
                return;

            cleanup();

            var time = TimeRange.ABSOLUTE.time(gst.call(GstMediaPlayer.this::pos), TimeUnit.NANOSECOND);
            segment(new Segment(time, range(), rate(), playing()));

            super.seek(segment);
        }

        @Override
        public void close() {
            cleanup();
            super.close();
        }

        @Override public String toString() { return "Playing{}"; }
    }

    class Seeking extends PlayerState {
        private Segment target;

        private final CountDownLatch latch;

        private final AtomicBoolean done = new AtomicBoolean(false);
        private final Bus.ASYNC_DONE doneListener;

        private final ArrayList<Bus.STATE_CHANGED> mediaListeners = new ArrayList<>();

        public Seeking(Segment segment, Segment target) {
            super(segment);
            this.target = target;

            var state = gst.call(pipeline::getState);
            expect(state == State.PAUSED || state == State.PLAYING, "Seek in invalid pipeline state " + state);

            System.out.println("M: SEEKING STATE=" + state);

            segment(target);

            latch = new CountDownLatch(media.size());
            doneListener = new Bus.ASYNC_DONE() {
                @Override
                public void asyncDone(GstObject source) {
                    if (latch.getCount() != 0 || !done.compareAndSet(false, true))
                        return;

                    System.out.println("M: SEEKING ASYNC_DONE LATCH=" + latch.getCount());

                    bus.disconnect(this);

                    transitionAsync(Seeking.this, () -> Seeking.this.target.playing
                        ? new Playing(target)
                        : new Paused(target));
                }
            };
            bus.connect(doneListener);

            long pos0 = target.time.absolute().max(start()).value();
            long pos1 = target.range.t1.absolute().value();

            System.out.println("M: SEEKING POS " + pos0 + " -> " + pos1);

            for (var m : media) {
                var sink = m.sink();

                var listener = new Bus.STATE_CHANGED() {
                    @Override
                    public void stateChanged(GstObject source, State old, State current, State pending) {
                        if (source != sink || pending != State.VOID_PENDING)
                            return;

                        pipeline.getBus().disconnect(this);
                        latch.countDown();

                        System.out.println("M: SEEK COUNTDOWN " + m.uri() + " IN " + current + " -> " + pending);

                        if (latch.getCount() == 0 && playing()) {
                            System.out.println("M: ASYNC_DONE OVERRIDE");
                            Seeking.this.doneListener.asyncDone(null);
                        }
                    }
                };
                mediaListeners.add(listener);
                bus.connect(listener);

                var range = m.range();
                var duration = m.duration();
                var offset = offsets.get(m).absolute();

                var end = duration;
                var pad = duration.subtract(new Duration(1, range.unit()));

                System.out.println("M: CLAMP " + m.uri());
                System.out.println("M: CLAMP " + duration + " (" + duration.absolute() + ") (+" + offset + ")");
                System.out.println("M: CLAMP END " + end + " (" + end.absolute() + ")");
                System.out.println("M: CLAMP PAD " + pad + " (" + pad.absolute() + ")");
//                System.out.println("M: CLAMP ABS " + start() + " -> " + end());

                var t0 = pos0 - offset;
                var t1 = pos1 - offset;
                var ofs = t0 < 0 ? Math.abs(t0) : 0;

                if (t1 >= pad.absolute())
                    t1 = end.absolute();

                t0 = Math.max(0, Math.min(pad.absolute(), t0));
                t1 = Math.max(0, Math.min(end.absolute(), t1));

                var flags = EnumSet.of(SeekFlags.FLUSH, SeekFlags.ACCURATE, SeekFlags.TRICKMODE);
                var event = new SeekEvent(target.rate, Format.TIME, flags, SeekType.SET, t0, SeekType.SET, t1);

                System.out.println("M: SEEKING MEDIA " + m.uri() + ": " + event.getStart() + " -> " + event.getStop() + " (+" + ofs + ")");

                gst.run(() -> {
                    sink.set("ts-offset", ofs);
                    sink.sendEvent(event);
                });
            }
        }

        private void cleanup() {
            if (done.compareAndSet(false, true))
                pipeline.getBus().disconnect(doneListener);

            mediaListeners.forEach(bus::disconnect);
        }

        @Override public void play() { target = new Segment(target.time, target.range, target.rate, true); }
        @Override public void pause() { target = new Segment(target.time, target.range, target.rate, false); }

        @Override
        public void seek(Segment segment) {
            if (done.get())
                return;

            cleanup();
            transition(this, () -> new Seeking(segment(), segment));
        }

        @Override public void close() {
            if (done.get())
                return;

            cleanup();
            transition(this, NotReady::new);
        }

        @Override public String toString() { return "Seeking{" + target + "}"; }
    }

//    class Stopped extends PlayerState {
//        private final Time time;
//        private final double rate;
//        private final Interval segment;
//
//        public Stopped(Time time, double rate, Interval segment) {
//            this.time = time;
//            this.rate = rate;
//            this.segment = segment;
//
//            System.out.println("M: STOPPED");
//        }
//
//        @Override public void play() {
//            transition(this, () -> new Seeking(time(), rate(), segment(), true, time().to(end()), rate()));
//        }
//
//        @Override public void pause() { /* do nothing */ }
//
//        @Override
//        public void seek(Interval segment, double rate) {
//            transition(this, () -> new Seeking(time(), rate(), segment(), true, false, segment, rate));
//        }
//
//        @Override
//        public void close() {
//            transition(this, NotReady::new);
//        }
//
//        @Override public boolean playing() { return false; }
//        @Override public boolean stopped() { return true; }
//        @Override public Time time() { return time; }
//        @Override public Interval segment() { return segment; }
//        @Override public double rate() { return rate; }
//
//        @Override public String toString() { return "Stopped{time=" + time() + "}"; }
//    }


    // Utility Methods //


    private void trace(Throwable e) {
        String DIVIDER = "====================================================================================";

        System.err.println(DIVIDER);
        e.printStackTrace();
        System.err.println(DIVIDER);
        for (Thread t : Thread.getAllStackTraces().keySet()) {
            if (t.getName().contains("gstreamer service thread")) {
                System.err.println("THREAD: " + t.getName());
                for (var el : t.getStackTrace()) {
                    System.err.println("    at " + el);
                }
                System.out.println(DIVIDER);
            }

            if (t.getName().contains("media player")) {
                System.err.println("THREAD: " + t.getName());
                for (var el : t.getStackTrace()) {
                    System.err.println("    at " + el);
                }
                System.out.println(DIVIDER);
            }
        }

        System.out.println("PLAYER SEGMENT: " + state.segment());
        System.out.println("PLAYER STATE: " + state);
        System.err.println(DIVIDER);
    }

    private volatile boolean errored = false;
    private RuntimeException error(Throwable e) {
        System.err.println("M: CRITICAL ERROR");
        trace(e);

        if (!errored) {
            JOptionPane.showMessageDialog(
                null,
                "Media player encountered an internal error!" + "\n\nApologies for the inconvenience. Please restart SignStream.",
                "Critical Media Player Error",
                JOptionPane.ERROR_MESSAGE);
            errored = true;
        }

        return new RuntimeException("GStreamer Error");
    }

    protected void expect(boolean condition, String message) {
        if (!condition)
            throw error(new AssertionError(message));
    }

    private void await(CountDownLatch latch) {
        try {
            expect(latch.await(5, java.util.concurrent.TimeUnit.SECONDS),
                "Timed out waiting for latch!");
        } catch (InterruptedException e) {
            throw error(e);
        }
    }

    private <V> V await(Future<V> future) {
        try {
            return future.get(5, java.util.concurrent.TimeUnit.SECONDS);
        } catch (Exception e) {
            throw error(e);
        }
    }

    private volatile boolean transitioning = false;
    private Future<Void> transitionAsync(PlayerState curr, Callable<PlayerState> fn) {
        return exec.runAsync(() -> {
            expect(!transitioning, "Nested transitions!");
            expect(curr != prevState, "Double transition!");

            try {
                System.out.println("M: BEGIN TRANSITION " + state.segment());
                System.out.println("M: BEGIN TRANSITION in " + state);
                transitioning = true;

                prevState = state;
                state = fn.call();

                transitioning = false;
                System.out.println("M: END TRANSITION in " + state);
                System.out.println("M: END TRANSITION " + state.segment());
            } catch (Exception e) {
                System.out.println("M: TRANSITION FAILED in " + state);
                throw error(e);
            }
        });
    }

    private void transition(PlayerState curr, Callable<PlayerState> fn) {
        await(transitionAsync(curr, fn));
    }

}
