package edu.bu.signstream.media.ui.painter;

import edu.bu.signstream.common.SSMath;
import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.media.data.Samples;
import edu.bu.signstream.media.data.Window;
import edu.bu.signstream.media.ui.color.ColorMap;

import javax.swing.*;
import java.awt.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class PitchPainter extends SamplesPainter {
    // Detection frequency range
    public static final double f0 = 75;
    public static final double f1 = 500;

    // Display frequency range
    public static final double df0 = f0;
    public static final double df1 = f1;

    // Detection period
//    private static final int period = 1024;
//    private static final int window = period / 4;
    private static final double periodsPerWindow = 3.0;
    private static final double dt = periodsPerWindow / f0 / 4.0;

    // Candidate thresholds
    private static final int maxnCandidates = Math.max(15, (int)Math.floor(f1 / f0));
    private static final double silenceThreshold = 0.03;
    private static final double voicingThreshold = 0.45;

    // Path finding costs
    private static final double octaveCost = 0.01;
    private static final double octaveJumpCost = 0.35;
    private static final double voicedUnvoicedCost = 0.14;

    static class Candidate {
        public final double freq;
        public final double strength;
        public final long sample;
        public Candidate(double freq, double strength, long sample) {
            this.freq = freq;
            this.strength = strength;
            this.sample = sample;
        }
    }
    private final ArrayList<Candidate> pitches = new ArrayList<>();


    public PitchPainter(Samples samples, ColorMap colors, TimeRange range, Duration page) {
        super(samples, colors, range, page);

        var start = System.currentTimeMillis();

        double dx = 1.0 / samples.rate;
        int nx = samples.length;

        double duration = dx * nx;
        if (f0 < periodsPerWindow / duration)
            throw new AssertionError();

        // Period samples
        int nsamp_period = (int)Math.floor(1.0 / dx / f0);
//        int halfnsamp_period = nsamp_period / 2 + 1;

        // window samples
        double dt_win = periodsPerWindow / f0;
        int nsamp_win = (int)Math.floor(dt_win / dx);
        int halfnsamp_win = nsamp_win / 2 - 1;
        nsamp_win = halfnsamp_win * 2;

        // Get the next highest power of 2 above the window size, since 2^n FFTs are faster
        int nsamp_fft = 1;
        while (nsamp_fft < nsamp_win)
            nsamp_fft *= 2;

        int frames = (int)Math.floor((duration - dt_win) / dt) + 1;
        if (frames < 1)
            throw new AssertionError();

        double mid = 0 - 0.5 * dx + 0.5 * duration;
        double dur = frames * dt;
        double t1 = mid - 0.5 * dur + 0.5 * dt;

        double globalMean = samples.avg();
        double globalPeak = samples.fold((a, b) ->
            Math.max(Math.abs(a - globalMean), Math.abs(b - globalMean)));

        var candidates = new ArrayList<ArrayList<Candidate>>(frames);
        for (int i = 0; i < frames; i++)
            candidates.add(null);

        var intensities = new double[frames];

        for (int frame = 0; frame < frames; frame++) {
            double t = t1 + frame * dt;

            int leftSample = (int)Math.floor(t / dx + 1);
            int rightSample = leftSample + 1;

            // Compute local mean in +/- period
            int startSample = rightSample - nsamp_period;
            int endSample = leftSample + nsamp_period;
            if (startSample < 0 || endSample > nx)
                throw new AssertionError();
            var localMean = samples.avg(startSample, endSample);

            // Center the samples in the window
            startSample = rightSample - halfnsamp_win;
            endSample = leftSample + halfnsamp_win;
            if (startSample < 0 || endSample > nx)
                throw new AssertionError();

            // Compute local peak +/- half period
//            startSample = Math.max(0, halfnsamp_win + 1 - halfnsamp_period);
//            endSample = Math.min(nsamp_win - 1, halfnsamp_win + halfnsamp_period);
//            if ((startSample = halfnsamp_win + 1 - halfnsamp_period) < 1)
//                startSample = 1;
//            if ((endSample = halfnsamp_win + halfnsamp_period) > nsamp_win)
//                endSample = nsamp_win;
            var localPeak = samples.ceil(startSample, endSample);
            intensities[frame] = localPeak > globalPeak ? 1.0 : localPeak / globalPeak;

            var centered = new Samples(nsamp_fft, samples.rate);
            for (int i = 0; i < nsamp_win; i++)
                centered.set(i, samples.get(i + startSample) - localMean);

            // Compute FFT autocorrelation
//            var ac = scaledSamples.autoCorrelate();
            var ac = centered.fftForward(Window.GAUSSIAN);
            for (int i = 0; i < ac.length; i++)
                ac.set(i, ac.get(i) * ac.get(i));
            ac = ac.fftBackward(Window.SQUARE);

            // Make a list of candidates for this frame
            var frameCandidates = new ArrayList<Candidate>();
            candidates.set(frame, frameCandidates);

            // Add the unvoiced (0 strength) candidate by default
            frameCandidates.add(new Candidate(0, 0, leftSample));

            var max = ac.get(0);
            for (int i = 1; i < ac.length - 1; i++) {
                var vprev = ac.get(i - 1) / max;
                var v = ac.get(i) / max;
                var vnext = ac.get(i + 1) / max;

                if (v >= vprev && v >= vnext && v > voicingThreshold) {
                    double dr = 0.5 * (vnext - vprev);
                    double d2r = 2.0 * v - vprev - vnext;

                    double freq = 1.0 / dx / (i + dr / d2r);
                    double strength = v + 0.5 * dr * dr / d2r;

                    var candidate = new Candidate(freq, strength, leftSample);
                    if (frameCandidates.size() < maxnCandidates) {
                        // Add the candidate immediately if there's space
                        frameCandidates.add(candidate);
                    } else {
                        // Otherwise find a weaker one and replace it
                        double weakest = 2.0;
                        int idx = 0;
                        for (int j = 1; j < maxnCandidates; j++) {
                            var c = frameCandidates.get(j);
                            double localStrength = c.strength - octaveCost * (Math.log(f1 / c.freq) / Math.log(2));
                            if (localStrength < weakest) {
                                weakest = localStrength;
                                idx = j;
                            }
                        }

                        if (idx != 0 && strength > weakest) {
                            frameCandidates.set(idx, candidate);
                        }
                    }
                }
            }
        }

        double timeStepCorrection = dt / dx;
        var octaveJumpCost = PitchPainter.octaveJumpCost * timeStepCorrection;
        var voicedUnvoicedCost = PitchPainter.voicedUnvoicedCost * timeStepCorrection;

        var delta = new Samples[frames];
        var psi = new int[frames][];

        for (int frame = 0; frame < frames; frame++) {
            delta[frame] = new Samples(maxnCandidates, samples.rate);
            psi[frame] = new int[maxnCandidates];

            double unvoicedStrength = 2 - intensities[frame] / (silenceThreshold / (1.0 + voicingThreshold));
            var frameCandidates = candidates.get(frame);
            for (int i = 0; i < frameCandidates.size(); i++) {
                var c = frameCandidates.get(i);
                if (c.freq >= f0 && c.freq <= f1) {
                    delta[frame].set(i, c.strength - octaveCost * (Math.log(f1 / c.freq) / Math.log(2)));
                } else {
                    delta[frame].set(i, unvoicedStrength);
                }
            }
        }

        double max;
        int place;
        for (int frame = 1; frame < frames; frame++) {
            var prevDelta = delta[frame - 1];
            var currDelta = delta[frame];
            var currPsi = psi[frame];

            var frameCandidates = candidates.get(frame);
            for (int i = 0; i < frameCandidates.size(); i++) {
                var ci = frameCandidates.get(i);
                double fi = ci.freq;

                max = -1e30;
                place = 0;

                var prevFrameCandidates = candidates.get(frame - 1);
                for (int j = 0; j < prevFrameCandidates.size(); j++) {
                    var cj = prevFrameCandidates.get(j);
                    double fj = cj.freq;

                    double transitCost;
                    boolean currVoiceless = !(fi >= f0 && fi <= f1);
                    boolean prevVoiceless = !(fj >= f0 && fj <= f1);

                    if (currVoiceless)
                        transitCost = prevVoiceless ? 0.0 : voicedUnvoicedCost;
                    else
                        transitCost = prevVoiceless
                            ? voicedUnvoicedCost
                            : octaveJumpCost * Math.abs(Math.log(fj / fi) / Math.log(2));

                    double cost = prevDelta.get(j) - transitCost + currDelta.get(i);
                    if (cost > max) {
                        max = cost;
                        place = j;
                    }
                }

                currDelta.set(i, max);
                currPsi[i] = place;
            }
        }

        int end = frames - 1;
        place = 0;
        max = delta[end].get(place);
        var lastCandidates = candidates.get(end);
        for (int i = 0; i < lastCandidates.size(); i++) {
            if (delta[end].get(i) > max) {
                place = i;
                max = delta[end].get(place);
            }
        }

        // Collect most likely pitches for each frame
        for (int frame = end; frame >= 0; frame--) {
            pitches.add(candidates.get(frame).get(place));
            place = psi[frame][place];
        }
        Collections.reverse(pitches);

        System.out.println("P: Calculated pitches in " + (System.currentTimeMillis() - start) + " ms");
    }

    @Override
    protected void paint(Graphics g, int w, int h, int s0, int s1) {
        // Plot candidates for each window
        g.setColor(colors.map(1));
        for (var p : pitches) {
            var s = p.sample;
            if (p.freq == 0 || s < s0 || s > s1) continue;

            var x = (int)SSMath.project(0, w, s0, s1, s);
            var y = (int)SSMath.project(h, 0, df0, df1, p.freq);
            g.fillOval(x - 2, y - 2, 4, 4);
        }
    }

    @Override
    public List<String> text(int w, int h, int mx, int my) {
        var text = super.text(w, h, mx, my);

        var freq = SSMath.project(f1, f0, 0, h, my);
        text.add(String.format("Frequency: %,.2f Hz", freq));

        int s0 = (int) Math.round(page * pageSamples);
        int s1 = (int) Math.round((page + 1) * pageSamples);
        pitches.stream()
            .filter(p -> p.sample >= s0 && p.sample <= s1)
            .min(Comparator.comparing(p -> Math.abs(SSMath.project(0, w, s0, s1, p.sample) - mx)))
            .ifPresent(p -> {
                if (p.freq == 0)
                    text.add("Detected Pitch: 0 Hz (Unvoiced)");
                else
                    text.add(String.format("Detected Pitch: %,.2f Hz", p.freq));
            });

        return text;
    }

    public void paintNearest(JComponent parent, Graphics g, Color color) {
        if (!cursor) return;

        var w = parent.getWidth();
        var h = parent.getHeight();

        int s0 = (int) Math.round(page * pageSamples);
        int s1 = (int) Math.round((page + 1) * pageSamples);
        pitches.stream()
            .filter(p -> p.sample >= s0 && p.sample <= s1)
            .min(Comparator.comparing(p -> Math.abs(SSMath.project(0, w, s0, s1, p.sample) - mx)))
            .ifPresent(p -> {
                var x = (int)SSMath.project(0, w, s0, s1, p.sample);
                var y = (int)SSMath.project(h, 0, df0, df1, p.freq);
                g.setColor(color);
                g.fillOval(x - 3, y - 3, 6, 6);
            });
    }
}
